#!/usr/bin/env python

"""
Managing free/busy period collections using database representations.

Copyright (C) 2014, 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from imiptools.dates import format_datetime
from imiptools.freebusy.common import FreeBusyCollectionBase, SupportAttendee, \
                                      SupportExpires, from_string, to_string
from imiptools.sql import DatabaseOperations

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# Conversion functions.

def to_copy_string(s, encoding):

    """
    Encode 's' using 'encoding' as a string suitable for use in tabular data
    acceptable to the PostgreSQL COPY command with \N as null.
    """

    s = to_string(s, encoding)
    return s is None and "\\N" or s

def to_copy_file(records):

    """
    Encode the given 'records' and store them in a file-like object for use with
    a tabular import mechanism. Return the file-like object.
    """

    io = StringIO()
    for values in records:
        l = []
        for v in values:
            l.append(to_copy_string(v, "utf-8"))
        io.write("\t".join(l))
        io.write("\n")
    io.seek(0)
    return io

def quote_column(column):

    "Quote 'column' using the SQL keyword quoting notation."

    return '"%s"' % column



# Collection abstractions.

class FreeBusyDatabaseCollection(FreeBusyCollectionBase, DatabaseOperations):

    """
    An abstraction for a collection of free/busy periods stored in a database
    system.
    """

    def __init__(self, cursor, table_name, column_names=None, filter_values=None,
        mutable=True, paramstyle=None):

        """
        Initialise the collection with the given 'cursor' and with the
        'table_name', 'column_names' and 'filter_values' configuring the
        selection of data. If 'mutable' is indicated, the collection may be
        changed; otherwise, an exception will be raised.
        """

        FreeBusyCollectionBase.__init__(self, mutable)
        DatabaseOperations.__init__(self, column_names, filter_values, paramstyle)
        self.cursor = cursor
        self.table_name = table_name

        self.period_columns = self.period_class.period_columns

    # List emulation methods.

    def __nonzero__(self):
        return len(self) and True or False

    def __iter__(self):
        query, values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                })
        self.cursor.execute(query, values)
        return iter(map(lambda t: self.make_period(t), self.cursor.fetchall()))

    def __len__(self):
        query, values = self.get_query(
            "select count(*) from %(table)s :condition" % {
                "table" : self.table_name
                })
        self.cursor.execute(query, values)
        result = self.cursor.fetchone()
        return result and int(result[0]) or 0

    def __getitem__(self, i):
        return list(iter(self))[i]

    # Dictionary emulation methods (even though this is not a mapping).

    def clear(self):
        query, values = self.get_query(
            "delete from %(table)s :condition" % {
                "table" : self.table_name
                })
        self.cursor.execute(query, values)

    # Operations.

    def insert_period(self, period):

        "Insert the given 'period' into the collection."

        self._check_mutable()

        columns, values = self.period_columns, period.as_tuple(string_datetimes=True)

        query, values = self.get_query(
            "insert into %(table)s (:columns) values (:values)" % {
                "table" : self.table_name
                },
            columns, [to_string(v, "utf-8") for v in values])

        self.cursor.execute(query, values)

    def insert_periods(self, periods):

        "Insert the given 'periods' into the collection."

        if not hasattr(self.cursor, "copy_from"):
            return FreeBusyCollectionBase.insert_periods(self, periods)

        self._check_mutable()

        columns = self.merge_default_columns(self.period_columns)

        all_values = []
        for period in periods:
            all_values.append(self.merge_default_values(period.as_tuple(string_datetimes=True)))

        f = to_copy_file(all_values)

        # Copy from the file-like object to the table.

        self.cursor.copy_from(f, self.table_name, columns=map(quote_column, columns))

    def remove_periods(self, periods):

        "Remove the given 'periods' from the collection."

        self._check_mutable()

        for period in periods:
            values = period.as_tuple(string_datetimes=True)

            query, values = self.get_query(
                "delete from %(table)s :condition" % {
                    "table" : self.table_name
                    },
                self.period_columns, [to_string(v, "utf-8") for v in values])

            self.cursor.execute(query, values)

    def remove_event_periods(self, uid, recurrenceid=None, participant=None):

        """
        Remove from the collection all periods associated with 'uid' and
        'recurrenceid' (which if omitted causes the "parent" object's periods to
        be referenced).

        If 'participant' is specified, only remove periods for which the
        participant is given as attending.

        Return the removed periods.
        """

        self._check_mutable()

        columns, values = ["object_uid"], [uid]

        if recurrenceid:
            columns.append("object_recurrenceid")
            values.append(recurrenceid)
        else:
            columns.append("object_recurrenceid is null")

        if participant:
            columns.append("attendee")
            values.append(participant)

        query, _values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, _values)
        removed = self.cursor.fetchall()

        query, values = self.get_query(
            "delete from %(table)s :condition" % {
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

        return map(lambda t: self.make_period(t), removed)

    # Specific period removal when updating event details.

    remove_specific_event_periods = remove_event_periods

    def remove_additional_periods(self, uid, recurrenceids=None):

        """
        Remove from the collection all periods associated with 'uid' having a
        recurrence identifier indicating an additional or modified period.

        If 'recurrenceids' is specified, remove all periods associated with
        'uid' that do not have a recurrence identifier in the given list.

        Return the removed periods.
        """

        self._check_mutable()

        if not recurrenceids:
            columns, values = ["object_uid", "object_recurrenceid is not null"], [uid]
        else:
            columns, values = ["object_uid", "object_recurrenceid not in ?", "object_recurrenceid is not null"], [uid, tuple(recurrenceids)]

        query, _values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, _values)
        removed = self.cursor.fetchall()

        query, values = self.get_query(
            "delete from %(table)s :condition" % {
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

        return map(lambda t: self.make_period(t), removed)

    def remove_affected_period(self, uid, start, participant=None):

        """
        Remove from the collection the period associated with 'uid' that
        provides an occurrence starting at the given 'start' (provided by a
        recurrence identifier, converted to a datetime). A recurrence identifier
        is used to provide an alternative time period whilst also acting as a
        reference to the originally-defined occurrence.

        If 'participant' is specified, only remove periods for which the
        participant is given as attending.

        Return any removed period in a list.
        """

        self._check_mutable()

        start = format_datetime(start)

        columns, values = ["object_uid", "start", "object_recurrenceid is null"], [uid, start]

        if participant:
            columns.append("attendee")
            values.append(participant)

        query, _values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, _values)
        removed = self.cursor.fetchall()

        query, values = self.get_query(
            "delete from %(table)s :condition" % {
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

        return map(lambda t: self.make_period(t), removed)

    def periods_from(self, period):

        "Return the entries in the collection at or after 'period'."

        start = format_datetime(period.get_start_point())

        columns, values = [], []

        if start:
            columns.append("start >= ?")
            values.append(start)

        query, values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

        return map(lambda t: self.make_period(t), self.cursor.fetchall())

    def periods_until(self, period):

        "Return the entries in the collection before 'period'."

        end = format_datetime(period.get_end_point())

        columns, values = [], []

        if end:
            columns.append("start < ?")
            values.append(end)

        query, values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

        return map(lambda t: self.make_period(t), self.cursor.fetchall())

    def get_overlapping(self, periods):

        """
        Return the entries in the collection providing periods overlapping with
        the given sorted collection of 'periods'.
        """

        overlapping = set()

        for period in periods:
            columns, values = self._get_period_values(period)

            query, values = self.get_query(
                "select %(columns)s from %(table)s :condition" % {
                    "columns" : self.columnlist(self.period_columns),
                    "table" : self.table_name
                    },
                columns, values)

            self.cursor.execute(query, values)

            overlapping.update(map(lambda t: self.make_period(t), self.cursor.fetchall()))

        overlapping = list(overlapping)
        overlapping.sort()
        return overlapping

    def remove_overlapping(self, period):

        "Remove all periods overlapping with 'period' from the collection."

        self._check_mutable()

        columns, values = self._get_period_values(period)

        query, values = self.get_query(
            "delete from %(table)s :condition" % {
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

    def _get_period_values(self, period):

        start = format_datetime(period.get_start_point())
        end = format_datetime(period.get_end_point())

        columns, values = [], []

        if end:
            columns.append("start < ?")
            values.append(end)
        if start:
            columns.append("end > ?")
            values.append(start)

        return columns, values

class FreeBusyGroupDatabaseCollection(SupportAttendee, FreeBusyDatabaseCollection):

    "A collection of quota group free/busy objects."

    def remove_specific_event_periods(self, uid, recurrenceid=None, attendee=None):

        """
        Remove from the collection all periods associated with 'uid' and
        'recurrenceid' (which if omitted causes the "parent" object's periods to
        be referenced) and any 'attendee'.

        Return the removed periods.
        """

        self._check_mutable()

        columns, values = ["object_uid"], [uid]

        if recurrenceid:
            columns.append("object_recurrenceid")
            values.append(recurrenceid)
        else:
            columns.append("object_recurrenceid is null")

        if attendee:
            columns.append("attendee")
            values.append(attendee)
        else:
            columns.append("attendee is null")

        query, _values = self.get_query(
            "select %(columns)s from %(table)s :condition" % {
                "columns" : self.columnlist(self.period_columns),
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, _values)
        removed = self.cursor.fetchall()

        query, values = self.get_query(
            "delete from %(table)s :condition" % {
                "table" : self.table_name
                },
            columns, values)

        self.cursor.execute(query, values)

        return map(lambda t: self.make_period(t), removed)

class FreeBusyOffersDatabaseCollection(SupportExpires, FreeBusyDatabaseCollection):

    "A collection of offered free/busy objects."

    pass

# vim: tabstop=4 expandtab shiftwidth=4
